# class X:
#     pass
#
#
# class Y(X):
#     pass
#
#
# class Z(Y):
#     pass
#
#
# class W(Z):
#     pass
#
#
# my_var = Z()
# t = type(my_var)
# print(issubclass(t, W))
# print(issubclass(t, Z))
# print(issubclass(t, Y))
# print(issubclass(t, X))
#


# class X:
#     def __init__(self):
#         print('hi')
#
#
# class Y(X):
#     def __init__(self):
#         super().__init__()
#         print('hello')
#
#
# class Z(Y):
#     pass
#
#
# class W(Y):
#     def __init__(self):
#         super().__init__()
#         print('OK?')
#
#
# a, b, c, d = X(), Y(), Z(), W()
from random import randint
from typing import List
from src.sabapipeline import Event, EventHandler
from src.sabapipeline import TriggerableEventBatchSource

class MyEvent(Event):
    def __init__(self, x):
        self.x = x
    def __str__(self):
        return f'({self.x})'

class X(TriggerableEventBatchSource):
    def get_event_batch(self) -> List[Event]:
        a, b = randint(1, 10), randint(1, 10)
        return [MyEvent(i) for i in range(a, a+b)]

class XEventHandler(EventHandler):
    def handle_event(self, event_to_handle: Event):
        print(event_to_handle)


x = X()
x.event_handler = XEventHandler()
for i in range(100):
    x.get_event_generator_function()



from ...__private.utilities import *
from .store import DataStore, SimpleRAMDataStore


class PipelineConfig:
    def __init__(self,
                 source_triggerers_thread_count: int = 1,
                 data_store: DataStore = None
                 # ray_config=None
                 ):
        self.source_triggerers_thread_count = source_triggerers_thread_count
        self.data_store = get_not_none(data_store, SimpleRAMDataStore())



from src.sabapipeline.config import DataStore, SimpleRAMDataStore

data_store: DataStore = SimpleRAMDataStore()
my_data_store: DataStore = data_store.get_branch("ali:", value_generator=lambda x: f'value for {x}')
print(my_data_store["saeed"])
print(my_data_store["asghar"])
print(my_data_store["ahmad"])
b = 2




from ...__private.utilities import *


class DataStore(ABC):
    @abstractmethod
    def get_value(self, key: str) -> Optional[Any]:
        raise NotImplementedError()

    @abstractmethod
    def set_value(self, key: str, value: Any) -> None:
        raise NotImplementedError()

    @abstractmethod
    def contains(self, key: str) -> bool:
        raise NotImplementedError()

    def __getitem__(self, item: str) -> Optional[Any]:
        return self.get_value(item)

    def __setitem__(self, key: str, value: Any) -> None:
        self.set_value(key, value)

    def __contains__(self, item: str) -> bool:
        return self.contains(item)

    def get_branch(self, prefix: str,
                   value_generator: Callable[[str], Any] = None,
                   change_event_handlers: List[Callable[[str, Optional[Any], Any], None]] = None):
        return DataStoreBranch(self, prefix, value_generator, change_event_handlers)


class DataStoreBranch(DataStore):
    def __init__(self,
                 parent_data_store: DataStore,
                 prefix: str,
                 value_generator: Callable[[str], Any],
                 change_event_handlers: List[Callable[[str, Optional[Any], Any], None]]
                 ):
        super().__init__()
        self._parent_data_store = parent_data_store
        self._prefix = prefix
        self._value_generator = value_generator
        self._change_event_handlers = change_event_handlers

    def get_value(self, key: str) -> Optional[Any]:
        result = self._parent_data_store.get_value(self._prefix + key)
        if result is None and self._value_generator is not None:
            result = self._value_generator(key)
            self.set_value(key, result)
            if self._change_event_handlers is not None and len(self._change_event_handlers) > 0:
                for handler in self._change_event_handlers:
                    handler(key, None, result)
        return result

    def set_value(self, key: str, value: Any) -> None:
        if self._change_event_handlers is not None and len(self._change_event_handlers) > 0:
            previous_value = self.get_value(key)
        else:
            previous_value = None
        self._parent_data_store.set_value(self._prefix + key, value)
        if previous_value is not None:
            for handler in self._change_event_handlers:
                handler(key, previous_value, value)

    def contains(self, key: str) -> bool:
        return self._parent_data_store.contains(self._prefix + key)


class SimpleRAMDataStore(DataStore):
    def __init__(self):
        super().__init__()
        self._storage: Dict[str, Any] = dict()

    def get_value(self, key: str) -> Optional[Any]:
        return self._storage.get(key, None)

    def set_value(self, key: str, value: Any) -> None:
        self._storage[key] = value

    def contains(self, key: str) -> bool:
        return key in self._storage



from .utilities import *
from .concepts import EventHandler, Event
from .elements import EventListener


class EventBus(EventHandler):
    def __init__(self):
        self.listeners: Dict[Type, List[EventListener]] = dict()

    def add_listener(self, listener: EventListener, type_to_listen: Type) -> None:
        if type_to_listen not in self.listeners:
            self.listeners[type_to_listen] = [listener]
        else:
            self.listeners[type_to_listen].append(listener)

    def handle_event(self, event_to_handle: Event):
        event_type: Type = type(event_to_handle)
        for listed_type in self.listeners:
            if issubclass(event_type, listed_type):
                for event_listener in self.listeners[listed_type]:
                    event_listener.process_event(event_to_handle)
